/* LFN Functions (Win95/WinNT) 1.0.0
 *   Emulates syntax of Borland-C++ Equivalent SFN functions.
 *   Will automatically call SFN functions if LFN not available.
 *  Copyright (c) 1996 by Peter Johnson.  All rights reserved.
 *  This code may be used and modified as users see fit, as long
 *   as the original copyright is unaltered.
 *  E-mail Peter Johnson (locke@mcs.net) if you have any questions.
 */
#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <dir.h>
#include <stdlib.h>
#include <errno.h>

#include "lfn.h"

typedef struct WIN32_FIND_DATA_s
{
	unsigned long attrib;		/* attribute found */
	unsigned long ctime;		/* creation time */
	unsigned long ctime2;
	unsigned long atime;		/* last access time */
	unsigned long atime2;
	unsigned long mtime;		/* last write time */
	unsigned long mtime2;
	unsigned long sizehigh;		/* high word of size, in bytes */
	unsigned long size;		/* low word of size, in bytes */
	unsigned long res[2];		/* reserved */
	char name[260];			/* LFN */
	char altname[14];		/* SFN */
} WIN32_FIND_DATA;


static int lfnsupport(const char *);
static int lfnfindclose(struct ffblk_lfn *ffblk);

/* determines if lfn supported, does not check any other flags */
static int lfnsupport(const char *drivepath)
{
	union REGS regs;
	struct SREGS segregs;
	char rootpath[4];	/* 'C:\' or 'D:\', etc */
	char buf[260];
	char filesystemname[256];
	
	fnsplit(drivepath, rootpath, NULL, NULL, NULL);
	if(!rootpath)
	{
		getcwd(buf, 260);
		fnsplit(buf, rootpath, NULL, NULL, NULL);
	}
	strcat(rootpath, "\\");
	
	segread(&segregs);
	
	regs.x.ax = 0x71A0;	/* Get Volume Information */
	segregs.es = FP_SEG(&filesystemname);	/* file system name */
	regs.x.di = FP_OFF(&filesystemname);
	regs.x.cx = sizeof(filesystemname);
	segregs.ds = FP_SEG(&rootpath[0]);	/* root directory path */
	regs.x.dx = FP_OFF(&rootpath[0]);
	
	intdosx(&regs, &regs, &segregs);	/* DOS call */
	
	if(regs.x.cflag != 0)
		return 0;
	if(regs.x.bx & 0x4000)	/* 0x4000 = FS_LFN_APIS */
		return 1;
	
	return 0;
}

int lfn_findfirst(const char *pathname, struct ffblk_lfn *ffblk, int attrib)
{
	union REGS regs;
	struct SREGS segregs;
	struct ffblk oldffblk;
	int i;
	WIN32_FIND_DATA finddata;
	
	if(!lfnsupport(pathname)) /* no LFN support */
	{
		/* call SFN findfirst */
		if(findfirst(pathname, &oldffblk, attrib) == -1)
			return -1;
		/* copy data into LFN ffblk */
		ffblk->lfn_handle = 0;
		ffblk->ff_attrib = oldffblk.ff_attrib;
		ffblk->ff_ftime = oldffblk.ff_ftime;
		ffblk->ff_fdate = oldffblk.ff_fdate;
		ffblk->ff_fsize = oldffblk.ff_fsize;
		strcpy(ffblk->ff_name, oldffblk.ff_name);
		for(i=0;i<21;i++)
			ffblk->ff_reserved[i] = oldffblk.ff_reserved[i];
		return 0;
	}
	
	/* LFN findfirst */
	segread(&segregs);
	
	regs.x.ax = 0x714E;	/* find first file */
	regs.x.cx = attrib;	/* search attributes */
	segregs.ds = FP_SEG(pathname);	/* filespec */
	regs.x.dx = FP_OFF(pathname);
	segregs.es = FP_SEG(&finddata);	/* find_data struct */
	regs.x.di = FP_OFF(&finddata);
	regs.x.si = 1;			/* DOS date/time values */
	
	intdosx(&regs, &regs, &segregs);	/* call DOS */
	
	if(regs.x.cflag != 0)
		return regs.x.ax;
	
	ffblk->lfn_handle = regs.x.ax;
	ffblk->ff_attrib = finddata.attrib;
	ffblk->ff_fsize = finddata.size;
	strcpy(ffblk->ff_name, finddata.name);
	*(long *)(&ffblk->ff_ftime) = (long)(finddata.mtime);
	*(long *)(&ffblk->lfn_ctime) = (long)(finddata.ctime);
	*(long *)(&ffblk->lfn_atime) = (long)(finddata.atime);
	
	/* no wildcards? */
	if(!strchr(pathname,'*') && !strchr(pathname,'?'))
		lfnfindclose(ffblk);
	return 0;
}

int lfn_findnext(struct ffblk_lfn *ffblk)
{
	union REGS regs;
	struct SREGS segregs;
	struct ffblk oldffblk;
	int i;
	WIN32_FIND_DATA finddata;
	
	if(!(ffblk->lfn_handle))	/* SFN findnext */
	{
		oldffblk.ff_attrib = ffblk->ff_attrib;
		oldffblk.ff_ftime = ffblk->ff_ftime;
		oldffblk.ff_fdate = ffblk->ff_fdate;
		oldffblk.ff_fsize = ffblk->ff_fsize;
		strcpy(oldffblk.ff_name, ffblk->ff_name);
		for(i=0;i<21;i++)
			oldffblk.ff_reserved[i] = ffblk->ff_reserved[i];
		
		if(findnext(&oldffblk) == -1)
			return -1;
		
		ffblk->ff_attrib = oldffblk.ff_attrib;
		ffblk->ff_ftime = oldffblk.ff_ftime;
		ffblk->ff_fdate = oldffblk.ff_fdate;
		ffblk->ff_fsize = oldffblk.ff_fsize;
		strcpy(ffblk->ff_name, oldffblk.ff_name);
		for(i=0;i<21;i++)
			ffblk->ff_reserved[i] = oldffblk.ff_reserved[i];
		return 0;
	}
	
	/* LFN findnext */
	segread(&segregs);
	
	regs.x.ax = 0x714F;	/* find next file */
	regs.x.bx = ffblk->lfn_handle;	/* handle from find first file */
	segregs.es = FP_SEG(&finddata);	/* find_data struct */
	regs.x.di = FP_OFF(&finddata);
	regs.x.si = 1;			/* DOS datetime format */
	
	intdosx(&regs, &regs, &segregs);	/* call DOS */
	
	if(regs.x.cflag == ENMFILE)	/* done searching? */
		lfnfindclose(ffblk);
	
	if(regs.x.cflag != 0)	/* error? */
		return -1;
	
	ffblk->ff_attrib = finddata.attrib;
	ffblk->ff_fsize = finddata.size;
	strcpy(ffblk->ff_name, finddata.name);
	*(long *)(&ffblk->ff_ftime) = (long)(finddata.mtime);
	*(long *)(&ffblk->lfn_ctime) = (long)(finddata.ctime);
	*(long *)(&ffblk->lfn_atime) = (long)(finddata.atime);
	return 0;
}

static int lfnfindclose(struct ffblk_lfn *ffblk)
{
	union REGS regs;
	
	if(ffblk->lfn_handle)	/* LFN */
	{
		regs.x.ax = 0x71A1;	/* find close */
		regs.x.bx = ffblk->lfn_handle;
		intdos(&regs, &regs);	/* call DOS */
	}
	/* SFN doesn't need findclose */
	return 0;
}

int lfn_getshortpath(const char *longpath, char *shortpath)
{
	union REGS regs;
	struct SREGS segregs;
	
	if(!lfnsupport(longpath)) /* no LFN support */
	{
		strcpy(shortpath, longpath);
		return 0;
	}

	segread(&segregs);
	
	regs.x.ax = 0x7160;
	regs.x.cx = 1;		/* get short path name */
	segregs.ds = FP_SEG(longpath);	/* long pathname */
	regs.x.si = FP_OFF(longpath);
	segregs.es = FP_SEG(shortpath);	/* short pathname */
	regs.x.di = FP_OFF(shortpath);
	
	intdosx(&regs, &regs, &segregs);

	if(regs.x.cflag != 0)
		return regs.x.ax;
	return 0;
}

int lfn_getshortname(const char *longname, char *shortname)
{
	char temp[260];
	char temp2[260], temp3[260];
	int ret;
	
	ret = lfn_getshortpath(longname, temp);
	if(ret)
		return ret;
	fnsplit(temp, NULL, NULL, temp2, temp3);
	fnmerge(temp, NULL, NULL, temp2, temp3);
	strcpy(shortname, temp);
	return 0;
}

int lfn_getcurdir(int drive, char *directory)
{
	union REGS regs;
	struct SREGS segregs;
	char disk[4];
	if(drive==0)
		disk[0] = getdisk() + 65;
	else
		disk[0] = drive+64;
	disk[1] = ':'; disk[2] = '\\'; disk[3] = 0;
	
	if(!lfnsupport(disk)) /* no LFN support */
		return(getcurdir(drive, directory));

	segread(&segregs);
	
	regs.x.ax = 0x7147;	/* get current directory */
	regs.h.dl = drive;
	segregs.ds = FP_SEG(directory);	/* path */
	regs.x.si = FP_OFF(directory);
	
	intdosx(&regs, &regs, &segregs);
	
	if(regs.x.cflag != 0)
		return regs.x.ax;
	return 0;
}

char *lfn_getcwd(char *buf, int buflen)
{
	char disk[4], dir[260];
	char *temp;
	disk[0] = getdisk() + 65;
	disk[1] = ':'; disk[2] = '\\'; disk[3] = 0;
	
	if(!lfnsupport(disk)) /* no LFN support */
		return(getcwd(buf, buflen));
	
	if(lfn_getcurdir(0, &dir[0]))
		return NULL;
	if(!buf)
	{
		temp = (char *)malloc(buflen);
		strcpy(temp, disk);
		strncat(temp, dir, buflen-3);
		return temp;
	}
	else
	{
		strcpy(buf, disk);
		strncat(buf, dir, buflen-3);
	}
	return buf;
}

int lfn_rename(const char *oldname, const char *newname)
{
	union REGS regs;
	struct SREGS segregs;
	
	if(!lfnsupport(oldname))	/* no LFN support */
		return(rename(oldname, newname));
	
	segread(&segregs);
	
	regs.x.ax = 0x7156;	/* rename file */
	segregs.ds = FP_SEG(oldname);	/* old name */
	regs.x.dx = FP_OFF(oldname);
	segregs.es = FP_SEG(newname);	/* new name */
	regs.x.di = FP_OFF(newname);
	
	intdosx(&regs, &regs, &segregs);
	
	if(regs.x.cflag != 0)
		return regs.x.ax;
	return 0;
}

int lfn_chdir(const char *path)
{
	char temp[260];
	lfn_getshortpath(path, temp);
	return(chdir(temp));
}

int lfn_remove(const char *filename)
{
	char temp[260];
	lfn_getshortpath(filename, temp);
	return(remove(temp));
}


/* LFN_FOPEN/LFN_FCLOSE */

typedef struct lfn_file_s
{
	char *lfn;		/* Long filename */
	char *sfn;		/* Short filename (temporary) */
	FILE *sfnfile;		/* FILE * to short filename */
} LFN_FILE;

#define active(p)	((p)->sfnfile != NULL)	/* is entry used? */

#define NSTATIC	20	/* static entries */

/* initialize static data */
static LFN_FILE _lfnb[NSTATIC] =
{
	{NULL, NULL, NULL}, {NULL, NULL, NULL},
	{NULL, NULL, NULL}, {NULL, NULL, NULL},
	{NULL, NULL, NULL}, {NULL, NULL, NULL},
	{NULL, NULL, NULL}, {NULL, NULL, NULL},
	{NULL, NULL, NULL}, {NULL, NULL, NULL},
	{NULL, NULL, NULL}, {NULL, NULL, NULL},
	{NULL, NULL, NULL}, {NULL, NULL, NULL},
	{NULL, NULL, NULL}, {NULL, NULL, NULL},
	{NULL, NULL, NULL}, {NULL, NULL, NULL},
	{NULL, NULL, NULL}, {NULL, NULL, NULL}
};

static char		sbuf[NSTATIC];
static char *		_smalllfnbuf = sbuf;
static LFN_FILE **	lfnbglue;
static LFN_FILE **	endglue;

/* internal functions */
static LFN_FILE *_findlfnp(void);
static int _l_morefiles(void);
static void _lwalk(void (*function)(LFN_FILE *));
static int lfnclose(LFN_FILE *);


FILE *lfn_fopen(const char *filename, const char *mode)
{
	register LFN_FILE *lfnp;
	FILE *sfnp;
	char exists = 0;
	char tempname[260];
	char temp[260], tempdrive[3], tempdir[260];
	
	lfnp = _findlfnp();
	if (lfnp == NULL)
		return NULL;
	
	if(!lfnsupport(filename)) /* no LFN support */
	{
		sfnp = fopen(filename, mode);
		if(!sfnp)
			return NULL;
		lfnp->lfn = NULL;
		lfnp->sfn = NULL;
		lfnp->sfnfile = sfnp;
		return sfnp;
	}
	
	/* LFN fopen */
	tmpnam(tempname);
	lfn_getshortpath(filename, temp);
	fnsplit(temp, tempdrive, tempdir, NULL, NULL);
	temp[0] = 0;
	fnmerge(temp, tempdrive, tempdir, NULL, NULL);
	strcat(temp, tempname);
	
	switch(*mode)
	{
	case 'a':
		exists = 1;
	case 'r':
		exists = 1;
	case 'w':
		break;
	default:
		return NULL;
	}
	
	if(exists)
		lfn_rename(filename, temp);
	sfnp = fopen(temp, mode);
	if(!sfnp)
	{
		if(exists)
			lfn_rename(temp, filename);
		return NULL;
	}
	lfnp->lfn = strdup(filename);
	lfnp->sfn = strdup(temp);
	lfnp->sfnfile = sfnp;
	
	return sfnp;
}

static int lfnclose(LFN_FILE *lfnp)
{
	fclose(lfnp->sfnfile);	lfnp->sfnfile = NULL;
	if(lfnp->lfn)
	{
		lfn_rename(lfnp->sfn, lfnp->lfn);
		free(lfnp->sfn); lfnp->sfn = NULL;
		free(lfnp->lfn); lfnp->lfn = NULL;
	}
	return 0;
}

int lfn_fclose(FILE *file)
{
	register LFN_FILE **lfnv;
	register LFN_FILE *lp;
	
	if (lfnbglue == NULL)
	{
		for (lp = _lfnb; lp < &_lfnb[NSTATIC]; lp++)
			if (lp->sfnfile == file)
				lfnclose(lp);
	}
	else
	{
		for (lfnv = lfnbglue; lfnv < endglue; lfnv++)
			if (*lfnv && ((*lfnv)->sfnfile == file))
				lfnclose(*lfnv);
	}
	return 1;
}

int lfn_fcloseall(void)
{
	_lwalk((void (*)(LFN_FILE *))lfnclose);
	return 0;
}

static LFN_FILE *_findlfnp(void)
{
	register LFN_FILE **lfnv, *lfnp;
	
	if (lfnbglue == 0)
	{
		for(lfnp = _lfnb; lfnp < _lfnb + NSTATIC; lfnp++)
			if (!active(lfnp))
				return lfnp;
		
		if (_l_morefiles() == 0) {
			return NULL;
		}
	}
	
	lfnv = lfnbglue;
	while ((*lfnv != NULL) && active(*lfnv))
		if (++lfnv >= endglue)
			return NULL;
	
	if (*lfnv == NULL)
		*lfnv = (LFN_FILE *)calloc(1, sizeof **lfnv);
	
	return *lfnv;
}

static int _l_morefiles(void)
{
	register LFN_FILE **lfnv;
	register LFN_FILE *lp;
	int nfiles;
	
	nfiles = 10;	/* buffer 10 files at a time */
	
	lfnbglue = (LFN_FILE **)calloc(nfiles, sizeof *lfnbglue);
	if (lfnbglue == NULL)
		return 0;
	
	endglue = lfnbglue + nfiles;
	
	for (lp = _lfnb, lfnv = lfnbglue; lp < &_lfnb[NSTATIC]; )
		*lfnv++ = lp++;
	
	_smalllfnbuf = (char *)calloc(nfiles, sizeof(*_smalllfnbuf));
	return 1;
}

static void _lwalk(void (*function)(LFN_FILE *))
{
	register LFN_FILE **lfnv;
	register LFN_FILE *lp;
	
	if (lfnbglue == NULL)
	{
		for (lp = _lfnb; lp < &_lfnb[NSTATIC]; lp++)
			if (active(lp))
				(*function)(lp);
	}
	else
	{
		for (lfnv = lfnbglue; lfnv < endglue; lfnv++)
			if (*lfnv && active(*lfnv))
				(*function)(*lfnv);
	}
}
