/* kluStor v.3.0.1
 *  Copyright (c) 1997 by Peter Johnson (locke@mcs.net).
 *  Read Doc/copying.pj for usage information.
 *
 * File: kluStor.cpp
 * Desc: kluStor file routines
 */

/* the includes */
#include <memory.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <dirent.h>

#include "kluStor.h"
#include "kluCmp.h"	/* [de]compression */

static char *kluStorCopyright = "kluStor v.3.0.1 Copyright (c) 1995-97 by "
	"Peter Johnson (locke@mcs.net).";

/* static functions */
static bool WriteBytes(FILE *file, void *addr, unsigned long size);
static bool CopyBytes(FILE *dest, FILE *source, unsigned long size);
static bool Exists(const char *name);

/* kluDir constructor
 * Set up data structures
 */
kluDir::kluDir(bool EnableCachingData)
{
	// Allocate 1 entry
	DirPtr = (struct DirPtr_s *) calloc(1, sizeof(struct DirPtr_s));
	NumEntries = 0; NumAlloced = 1;
	CachingData = EnableCachingData;
	kluFrom = NULL;
	ExtBasePath = NULL;
}

/* kluDir::GetSubdir
 * Fill current kluDir with subdirectory of another
 */
void kluDir::GetSubdir(const kluDir &basedir, const char *subdir, bool EnableCachingData)
{
	kluFrom = basedir.kluFrom;	// copy klu file info
	ExtBasePath = strdup(basedir.ExtBasePath);	// duplicate base path

	if(!subdir)		// just do a copy
	{
		CachingData = basedir.CachingData;
		NumEntries = basedir.NumEntries; NumAlloced = basedir.NumAlloced;
		DirPtr = (struct DirPtr_s *) calloc(NumAlloced, sizeof(struct DirPtr_s));
		memcpy(DirPtr, basedir.DirPtr, NumEntries*sizeof(struct DirPtr_s));
 	}
	else				// get a subdir
	{
		// add subdir onto base path
		int len = strlen(subdir);
		ExtBasePath = (char *)realloc(ExtBasePath, strlen(ExtBasePath)+len+1);
		strcat(ExtBasePath, subdir);
		CachingData = EnableCachingData;
		NumEntries = 0; NumAlloced = 50;

		// add a slash
		char realsub[512];
		strcpy(realsub, subdir);
		strcat(realsub, "/"); len++;

		// alloc a base of 50 entries to begin with
		DirPtr = (struct DirPtr_s *) calloc(50, sizeof(struct DirPtr_s));
		// search through entire directory
		for(unsigned long i=0; i<basedir.NumEntries; i++)
			if(!strncmp(realsub, basedir.DirPtr[i].Saved.name, len))
			{
				if(NumEntries>=NumAlloced)
				{
					NumAlloced += 10;	// alloc more room
					DirPtr = (struct DirPtr_s *) realloc(DirPtr, NumAlloced*sizeof(struct DirPtr_s));
				}
				// copy found entry
				DirPtr[NumEntries] = basedir.DirPtr[i];
				// remove subdirectory from name
				memmove(&DirPtr[NumEntries].Saved.name[0], &DirPtr[NumEntries].Saved.name[len],
					512 - len);
				DirPtr[NumEntries].Memory.data = NULL;		// safety measure
				NumEntries++;
			}
	}
}

/* kluDir destructor
 * Free everything
 */
kluDir::~kluDir()
{
	if(!NumAlloced)
		return;
	if(ExtBasePath)
		free(ExtBasePath);
	// Free all cached data
	for(unsigned long i=0; i<NumEntries; i++)
		if(DirPtr[i].Memory.data)
			free(DirPtr[i].Memory.data);
	free(DirPtr);
}

/* kluDir::SetCachingData
 * If disabling cache, erase all cached data
 */
void kluDir::SetCachingData(bool cache)
{
	if(!cache && CachingData)
		for(unsigned long i=0; i<NumEntries; i++)
			if(DirPtr[i].Memory.data)
			  	free(DirPtr[i].Memory.data);
	CachingData = cache;
}

/* kluDir::Find
 * Find one entry
 */
unsigned long kluDir::Find(const char *name)
{
	char str[512];

	ClearError();

	int len = strlen(name);
	strcpy(str, name); strlwr(str);
	for(unsigned long i=0; i<NumEntries; i++)
		if(!strncmp(name, DirPtr[i].Saved.name, len))
			return i;
	SetError("entry %s not found", name);
	return 0xffffffff;
}

/* kluDir::Read
 * Read from entry, decompressing/caching as neccessary
 */
bool kluDir::Read(unsigned long index, void *data, long len, unsigned long offset)
{
	ClearError();
	if(index==0xffffffff)	// entry not found
		return false;

	if(DirPtr[index].Saved.comp_type > NumKluCompress)
		SetError("Invalid compression type in entry %s", DirPtr[index].Saved.name);

	// use cached data if possible
	if(DirPtr[index].Memory.data)
	{
		memcpy(data, (char *)DirPtr[index].Memory.data + offset,
			(len==-1)?DirPtr[index].Saved.decomp_size-offset:len);
	}
	else	// must read from file
	{
		unsigned long srclen = DirPtr[index].Saved.size, destlen = DirPtr[index].Saved.decomp_size;
		void *src = malloc(srclen), *dest = malloc(destlen);
		if(!src || !dest)
			return false;
		if(DirPtr[index].Memory.not_in_klu)	// in external file
		{
			char extfn[512];
			strcpy(extfn, ExtBasePath);
			strcat(extfn, DirPtr[index].Saved.name);
			FILE *ext = fopen(extfn, "rb");
			if(!ext)
				return false;
			fread(src, srclen, 1, ext);
			fclose(ext);
		}
		else	// read from klu file
		{
			fseek(kluFrom, DirPtr[index].Saved.start, SEEK_SET);
			fread(src, srclen, 1, kluFrom);
		}
		
		// Decompress
		kluCompress[DirPtr[index].Saved.comp_type].decomp(dest, destlen, src, srclen);

		memcpy(data, (char *)dest + offset, (len==-1)?destlen-offset:len);
		
		free(src);
		if(CachingData)	// Save in cache if enabled
			DirPtr[index].Memory.data = dest;
		else
			free(dest);
	}
	return true;
}

/* kluDir::Sort
 * Sort the klu directory
 */
static int SortCompare(const void *dir1, const void *dir2);

void kluDir::Sort(void)
{
	ClearError();

	qsort(DirPtr, NumEntries, sizeof(struct DirPtr_s), &SortCompare);
}

static int SortCompare(const void *dir1, const void *dir2)
{
	return strcmp(((struct kluDir::DirPtr_s *)(dir1))->Saved.name, ((struct kluDir::DirPtr_s *)(dir2))->Saved.name);
}

/* kluDir::Merge
 * Merge with another
 */
bool kluDir::Merge(const kluDir &base, const char *basedir)
{
	ClearError();
	char bdir[512];
	char cur[512];
	strcpy(bdir, basedir);
	strcat(bdir, "/");
	strlwr(bdir);
	int len = strlen(bdir);

	bool SlowMerge = false;
	for(unsigned long i=0; i<NumEntries; i++)
		if(!strncmp(DirPtr[i].Saved.name, bdir, len))
		{
			SlowMerge = true;	// must use slow merge if dup entries exist
			break;
		}

	if(SlowMerge)	// Yuck.. must use O(n^2) merge; TODO: Optimize!!
		for(unsigned long i=0; i<base.NumEntries; i++)
		{
			strcpy(cur, bdir); strcat(cur, base.DirPtr[i].Saved.name);
			len = strlen(cur);
			// See if entry already exists
			unsigned long found = 0xffffffff;
			for(unsigned long j=0; j<NumEntries; j++)
				if(!strncmp(cur, DirPtr[j].Saved.name, len))
				{
					found = j;
					break;
				}
			if(found==0xffffffff)	// Nope, just add it to end
			{
				found = NumEntries;
				if(found>=NumAlloced)
				{
					NumAlloced += 10;
					DirPtr = (struct DirPtr_s *) realloc(DirPtr, NumAlloced*sizeof(struct DirPtr_s));
				}
				NumEntries++;
			}
			DirPtr[found] = base.DirPtr[i];
		}
	else	// Good.. use O(n) merge, just tacking new entries onto end
	{
		NumAlloced += base.NumEntries;
		DirPtr = (struct DirPtr_s *) realloc(DirPtr, NumAlloced*sizeof(struct DirPtr_s));
		memcpy(&DirPtr[NumEntries], base.DirPtr, base.NumEntries*sizeof(struct DirPtr_s));
		NumEntries += base.NumEntries;
	}
	return true;
}

/* kluDir::Rename
 * Rename an entry
 */
bool kluDir::Rename(const char *name, const char *newname)
{
	unsigned long found = Find(name);	// Find entry
	if(found==0xffffffff)
		return false;
	memset(DirPtr[found].Saved.name, 0, 512);
	strcpy(DirPtr[found].Saved.name, newname);	// Rename it
	strlwr(DirPtr[found].Saved.name);
	return true;
}

/* kluDir::Delete
 * Delete an entry
 */
bool kluDir::Delete(const char *name)
{
	unsigned long found = Find(name);	// Find entry
	if(found==0xffffffff)
		return false;
	if(DirPtr[found].Memory.data)	// Erase cache, if any
		free(DirPtr[found].Memory.data);
	DirPtr[found].Saved.name[0] = 0;	// Delete entry's name
	return true;
}

/* kluDir::Add
 * Add an entry (from memory)
 */
bool kluDir::Add(const char *name, void *data, unsigned long size, unsigned char comp_type)
{
	char str[512];

	ClearError();

	unsigned long found = 0xffffffff;
	int len = strlen(name);
	strcpy(str, name); strlwr(str);
	// find an open or duplicate entry
	for(unsigned long i=0; i<NumEntries; i++)
	{
		if((DirPtr[i].Saved.name[0] == 0) && (found==0xffffffff))
			found = i;
		else if(!strncmp(str, DirPtr[i].Saved.name, len))
		{
			SetError("%s already exists.", str);	// do NOT replace
			return false;
		}
	}
	if(found==0xffffffff)	// Must append to end, so check space
	{
		found = NumEntries;
		if(found>=NumAlloced)
		{
			NumAlloced += 10;
			DirPtr = (struct DirPtr_s *) realloc(DirPtr, NumAlloced*sizeof(struct DirPtr_s));
		}
		NumEntries++;
	}
	// Set up structure
	DirPtr[found].Saved.start = 0;
	DirPtr[found].Saved.size = size;
	DirPtr[found].Saved.decomp_size = size;
	DirPtr[found].Saved.comp_type = comp_type;
	memset(DirPtr[found].Saved.name, 0, 512);
	strcpy(DirPtr[found].Saved.name, str);
	DirPtr[found].Memory.not_in_klu = true;
	DirPtr[found].Memory.data = data;	// Store in cache
	return true;
}

/* kluFile constructor
 * Open and load a klu file
 */
kluFile::kluFile(const char *name, bool ScanExtDir)
{
	unsigned long dirsize, dirstart, dir_comp_size;
	unsigned char dir_comp_type;
	char *ptr;

	ClearError();

	filename = strdup(name);
	// Set up base path
	dir.ExtBasePath = strdup(filename);
	ptr = strrchr(dir.ExtBasePath, '.'); ptr[0] = '/'; ptr[1] = 0;
	dir.CachingData = false;	// Default: off

	if(!Exists(name))	// create a new file
	{
		fileinfo = NULL;
		dir.kluFrom = NULL;
		type[0] = 'k'; type[1] = 'l'; type[2] = 'u'; type[3] = '3';
	}
	else
	{
		// open the klu file
		if ((fileinfo = fopen(filename, "rb")) == NULL)
		{
			SetError("\"%s\" does not exist", filename);
			return;
		}

		// read in the klu directory info
		if(!BasicRead(type, 4)) return;
		if (strncmp(type, "klu3", 4))
		{
			SetError("\"%s\" is not a valid klu 3.0 file", filename);
			return;
		}
		if(!BasicRead(&dirsize, sizeof(unsigned long))) return;
		if(!BasicRead(&dirstart, sizeof(unsigned long))) return;
		if(!BasicRead(&dir_comp_size, sizeof(unsigned long))) return;
		if(!BasicRead(&dir_comp_type, sizeof(unsigned char))) return;

		// read in the klu directory itself
		dir.kluFrom = fileinfo;
		dir.NumEntries = dirsize; dir.NumAlloced = dirsize;
		dir.DirPtr = (struct kluDir::DirPtr_s *) calloc(dirsize, sizeof(struct kluDir::DirPtr_s));
		struct kluDir::DirPtr_s::Saved_s *uncomp = (struct kluDir::DirPtr_s::Saved_s *)
			malloc(sizeof(struct kluDir::DirPtr_s::Saved_s)*dirsize);
		ptr = (char *) malloc(dir_comp_size);
		if(!BasicSeek(dirstart)) return;
		if(!BasicRead(ptr, dir_comp_size)) return;
		// decompress
		kluCompress[dir_comp_type].decomp(uncomp, sizeof(struct kluDir::DirPtr_s::Saved_s) * dirsize,
			ptr, dir_comp_size);
		free(ptr);
		for(unsigned long i=0; i<dirsize; i++)
			dir.DirPtr[i].Saved = uncomp[i];
		free(uncomp);
	}

	// read subdirectory structure
	if(ScanExtDir)
		ReadSubdir(dir.ExtBasePath);
}

/* kluFile::ReadSubdir
 * Scan subdirectory(s) for additions to klu
 */
bool kluFile::ReadSubdir(char *basepath)
{
	struct dirent *ent;
	DIR *basedir = opendir(basepath);
	struct stat st;
	struct kluDir::DirPtr_s kluent;
	char		*name;

	if(!basedir)
		return true;
	
	// init constant parts of klu entry structure
	kluent.Memory.not_in_klu = true; kluent.Memory.data = NULL;
	kluent.Saved.start = 0;

	while((ent = readdir(basedir)) != NULL)
	{
		name = ent->d_name;

		if (name[0] == '.')	// discard . and ..
			continue;
		memset(kluent.Saved.name, 0, 512);
		strcpy(kluent.Saved.name, basepath);
		strcat(kluent.Saved.name, name);
		strlwr(kluent.Saved.name);

		if (stat(kluent.Saved.name, &st) == -1)
		{
			SetError("fstating %s", kluent.Saved.name);
			return false;
		}
		if (st.st_mode & S_IFDIR)	// directory?
		{
			strcat(kluent.Saved.name, "/");
			if(!ReadSubdir(kluent.Saved.name))	// recurse
				return false;
			continue;
		}

		// init vital info
		int len = strlen(dir.ExtBasePath);
		memmove(&kluent.Saved.name[0], &kluent.Saved.name[len],
			512 - len);
		kluent.Saved.size = st.st_size;
		kluent.Saved.decomp_size = st.st_size;
		kluent.Saved.comp_type = 0;	// default: no compression

		// actually write into kluDir structure
		unsigned long index = Find(kluent.Saved.name);
		index = index==0xffffffff?dir.NumEntries:index;
		if(dir.NumEntries>=dir.NumAlloced)
		{
			dir.NumAlloced += 10;
			dir.DirPtr = (struct kluDir::DirPtr_s *) realloc(dir.DirPtr,
				dir.NumAlloced*sizeof(struct kluDir::DirPtr_s));
		}
		dir.DirPtr[index] = kluent;
		if(index==dir.NumEntries)
			dir.NumEntries++;
	}

	closedir(basedir);
	return true;
}

/* kluFile destructor
 * Free alloced memory
 */
kluFile::~kluFile()
{
	if(filename)
		free(filename);
	if(fileinfo)
		fclose(fileinfo);
}

/* kluFile::BasicRead
 * Read bytes from a file and store it into an address with error checking
 */
bool kluFile::BasicRead(void *addr, unsigned long size)
{
	if (fread(addr, 1, size, fileinfo) != size)
	{
		SetError("error reading from \"%s\"", filename);
		return false;
	}
	return true;
}

/* kluFile::BasicSeek
 * Go to offset of klu file with error checking
 */
bool kluFile::BasicSeek(unsigned long offset, FILE *f)
{
	FILE *f2 = f?f:fileinfo;
	if (fseek(f2, offset, 0))
	{
		SetError("error seeking in file");
		return false;
	}
	return true;
}

/* kluFile::Save
 * Save klu to disk, including external files and in-memory data if wanted
 */
bool kluFile::Save(const char *name, bool SaveNotInKlu, bool SaveInMemory,
	unsigned char dir_comp_type, bool (*Progress)(const char *curname,
	unsigned long num, unsigned long totnum, unsigned long pos))
{
	FILE *save;
	unsigned long dirstart = 0, dir_comp_size = 0;
	char extfn[512];
	struct kluDir::DirPtr_s::Saved_s *final;
	unsigned long finalAlloced = 50, finalEntries = 0;
	char *ptr;

	ClearError();

	if ((save = fopen(name, "wb")) == NULL)
	{
		SetError("cannot create \"%s\"", filename);
		return false;
	}

	// write header
	if(!WriteBytes(save, type, 4)) return false;
	if(!WriteBytes(save, &dir.NumEntries, sizeof(unsigned long))) return false;
	if(!WriteBytes(save, &dirstart, sizeof(unsigned long))) return false;
	if(!WriteBytes(save, &dir_comp_size, sizeof(unsigned long))) return false;
	if(!WriteBytes(save, &dir_comp_type, sizeof(unsigned char))) return false;
	dirstart += 4+3*sizeof(unsigned long)+sizeof(unsigned char);

	// write/copy entries
	final = (struct kluDir::DirPtr_s::Saved_s *) calloc(50, sizeof(struct kluDir::DirPtr_s::Saved_s));

	bool addit = true;
	for(unsigned long i=0; i<dir.NumEntries; i++, addit = true)
	{
		if(Progress)	// Progress callback
			if(!Progress(dir.DirPtr[i].Saved.name, i+1, dir.NumEntries, dirstart - 17))
			{
				SetError("Aborted");
				free(final);
				fclose(save);
				return false;
			}
		if(!dir.DirPtr[i].Saved.name[0])	// Deleted entry
			continue;
		if(dir.DirPtr[i].Memory.not_in_klu)	// memory or external
		{
			strcpy(extfn, dir.ExtBasePath);
			strcat(extfn, dir.DirPtr[i].Saved.name);
			FILE *ext = fopen(extfn, "rb");	// try to open
			
			if(!ext && !dir.DirPtr[i].Memory.data)
				return false;	// uh oh

			unsigned long decompsize = dir.DirPtr[i].Saved.decomp_size;
			unsigned long compsize = decompsize * 1.001 + 1024;
			void *inbuf, *outbuf = malloc(compsize);
			if(ext)	// read from file
			{
				addit = SaveNotInKlu;
				inbuf = malloc(decompsize);

				if (fread(inbuf, 1, decompsize, ext) != decompsize)
					return false;
			}
			else	// read from memory
			{
				addit = SaveInMemory;
				inbuf = dir.DirPtr[i].Memory.data;
			}
			if(addit)	// save it
			{
				dir.DirPtr[i].Saved.comp_type = dir_comp_type;
				// compress entry
				kluCompress[dir_comp_type].comp(outbuf, &compsize, inbuf,
					decompsize);

				if(!WriteBytes(save, outbuf, compsize)) return false;
				dir.DirPtr[i].Saved.size = compsize;
			}

			if(ext)
			{
				free(inbuf);
				fclose(ext);
			}
			free(outbuf);
		}
		else	// copy from old to new klu
		{
			BasicSeek(dir.DirPtr[i].Saved.start);
			CopyBytes(save, fileinfo, dir.DirPtr[i].Saved.size);
		}
		if(addit)	// add to saved directory
		{
			if(finalEntries>=finalAlloced)
			{
				finalAlloced += 10;
				final = (struct kluDir::DirPtr_s::Saved_s *)
					realloc(final, finalAlloced*sizeof(struct kluDir::DirPtr_s::Saved_s));
			}
			final[finalEntries] = dir.DirPtr[i].Saved;
			final[finalEntries].start = dirstart;
			finalEntries++;
		}
		dirstart += dir.DirPtr[i].Saved.size;	// update start index
	}

	// write directory
	dir_comp_size = finalEntries * sizeof(struct kluDir::DirPtr_s::Saved_s) * 1.001 + 1024;
	ptr = (char *) malloc(dir_comp_size);
	// compress it first
	kluCompress[dir_comp_type].comp(ptr, &dir_comp_size, final,
		sizeof(struct kluDir::DirPtr_s::Saved_s) * finalEntries);
	if(!WriteBytes(save, ptr, dir_comp_size)) return false;
	free(ptr);

	// update header
	if(!BasicSeek(4, save)) return false;
	if(!WriteBytes(save, &finalEntries, sizeof(unsigned long))) return false;
	if(!WriteBytes(save, &dirstart, sizeof(unsigned long))) return false;
	if(!WriteBytes(save, &dir_comp_size, sizeof(unsigned long))) return false;

	delete final;	// delete save directory

	fclose(save);	// close up
	return true;
}

/* WriteBytes
 * Output bytes to a binary file with error checking
 */
static bool WriteBytes(FILE *file, void *addr, unsigned long size)
{
	while (size > 0x8000)
	{
		if (fwrite(addr, 1, 0x8000, file) != 0x8000)
			return false;
		addr = (char *)addr + 0x8000;
		size -= 0x8000;
	}
	if (fwrite(addr, 1, size, file) != size)
		return false;
	return true;
}

/* CopyBytes
 * Copy bytes from a binary file to another with error checking
 */
static bool CopyBytes(FILE *dest, FILE *source, unsigned long size)
{
	void *data;

	data = malloc(0x8000 + 2);
	while (size > 0x8000)
	{
		if (fread(data, 1, 0x8000, source) != 0x8000)
			return false;
		if (fwrite(data, 1, 0x8000, dest) != 0x8000)
			return false;
		size -= 0x8000;
	}
	if (fread(data, 1, size, source) != size)
		return false;
	if (fwrite(data, 1, size, dest) != size)
		return false;
	free(data);
	return true;
}

/* Exists
 * Check if a file exists and is readable
 */
static bool Exists(const char *name)
{
	FILE *test;

	if ((test = fopen(name, "rb")) == NULL)
		return false;
	fclose(test);
	return true;
}
