/* kluStor v.3.0.1
 *  Copyright (c) 1997 by Peter Johnson (locke@mcs.net).
 *  Read Doc/copying.pj for usage information.
 *
 * File: kluStor.h
 * Desc: kluStor main header file
 */
#if !defined(KLUSTOR_H)
#define KLUSTOR_H

#include <stdio.h>
#include <string.h>

// class definitions
#include <stdarg.h>
class kluError
{
private:
	char ErrorStr[256];
protected:
	void ClearError(void)
	{
		ErrorStr[0] = 0;
	}
	void SetError(char *errstr, ...)
	{
		va_list args;
		va_start(args, errstr);
		vsprintf(ErrorStr, errstr, args);
		va_end(args);
	}
public:
	kluError()
	{
		ErrorStr[0] = 0;
	}
	char *GetError(void)
	{
		return(ErrorStr[0] ? ErrorStr : NULL);
	}
};

class kluFile;

class kluDir : public kluError
{
	friend kluFile;
private:
	struct DirPtr_s
	{
// make sure DJGPP packs this structure!
#ifdef __DJGPP__
#pragma pack(1)
#endif
		struct Saved_s	// saved to disk
		{
			char name[512];			// name of data block
			unsigned long start;		// offset to start of data
			unsigned long size;		// (compressed) byte size of data
			unsigned long decomp_size;	// decompressed byte size of data
			unsigned char comp_type;	// compression type (see klu.txt)
		} Saved;
#ifdef __DJGPP__
#pragma pack()
#endif
		struct Memory_s	// in memory only
		{
			bool not_in_klu;           // not in klu file (e.g. in external dir)
			void *data;                // decompressed data (if caching enabled)
		} Memory;
	} *DirPtr;
	unsigned long NumEntries, NumAlloced;
	bool CachingData;
	FILE *kluFrom;
	char *ExtBasePath;

	void GetSubdir(const kluDir &basedir, const char *subdir, bool ECD);	// get a subdir / copy constructor
public:
	kluDir(bool EnableCachingData = false);
	kluDir(const kluDir &basedir, const char *subdir = NULL, bool ECD = false)	// get a subdir / copy constructor
	{
		GetSubdir(basedir, subdir, ECD);
	}
	kluDir(const kluFile &base, const char *subdir = NULL, bool ECD = false);	// get a subdir
	~kluDir();

	bool IsCachingData(void)
	{
		return CachingData;
	}
	void SetCachingData(bool cache);		// if cache=false, clears cache; the next 2 fcns don't reset the cache
	void DisableCachingData(void)
	{
		CachingData = false;
	}
	void EnableCachingData(void)
	{
		CachingData = true;
	}

	unsigned long Find(const char *name);
	unsigned long GetSize(unsigned long index)
	{
		return((index==0xffffffff) ? 0 : DirPtr[index].Saved.decomp_size);
	}
	unsigned long GetSize(const char *name)
	{
		return GetSize(Find(name));
	}
	bool Read(unsigned long index, void *data, long len = -1, unsigned long offset = 0);
	bool Read(const char *name, void *data, long len = -1, unsigned long offset = 0)
	{
		return Read(Find(name), data, len, offset);
	}

	unsigned long GetNumEntries(void)
	{
		return NumEntries;
	}
	char *GetName(unsigned long index)
	{
		return DirPtr[index].Saved.name;
	}
	unsigned long GetCompSize(unsigned long index)
	{
		return DirPtr[index].Saved.size;
	}
	unsigned char GetCompType(unsigned long index)
	{
		return DirPtr[index].Saved.comp_type;
	}
	bool IsExternal(unsigned long index)
	{
		return DirPtr[index].Memory.not_in_klu;
	}

	bool Rename(const char *name, const char *newname);
	bool Delete(const char *name);
	bool Add(const char *name, void *data, unsigned long size, unsigned char comp_type);
	bool Merge(const kluDir &base, const char *basedir);
	void Sort(void);
};

class kluFile : public kluError
{
private:
	char *filename;	// name of the klu file
	FILE *fileinfo;	// C file stream information
	char type[4];		// magic string ("klu3")

	bool BasicRead(void *addr, unsigned long size);
	bool BasicSeek(unsigned long offset, FILE *f = NULL);

	bool ReadSubdir(char *basepath);
public:
	kluDir dir;

	kluFile(const char *name, bool ScanExtDir = true);
	~kluFile();
	bool Save(const char *name, bool SaveNotInKlu = true,
		bool SaveInMemory = true, unsigned char DirCompType = 0,
		bool (*Progress)(const char *curname, unsigned long num,
		unsigned long totnum, unsigned long pos) = NULL);

	// redirect standard directory ops to kluDir
	bool IsCachingData(void)
	{ return dir.CachingData; }
	void SetCachingData(bool cache)	// if cache=false, clears cache; the next 2 fcns don't reset the cache
	{ dir.SetCachingData(cache); }
	void DisableCachingData(void)
	{ dir.CachingData = false; }
	void EnableCachingData(void)
	{ dir.CachingData = true; }

	unsigned long Find(const char *name)
	{ return dir.Find(name); }
	unsigned long GetSize(unsigned long index)
	{ return dir.DirPtr[index].Saved.decomp_size; }
	unsigned long GetSize(const char *name)
	{ return GetSize(Find(name)); }
	bool Read(unsigned long index, void *data, long len = -1, unsigned long offset = 0)
	{ return dir.Read(index, data, len, offset); }
	bool Read(const char *name, void *data, long len = -1, unsigned long offset = 0)
	{ return dir.Read(Find(name), data, len, offset); }
};

inline kluDir::kluDir(const kluFile &base, const char *subdir, bool ECD)	// get a subdir
{
	GetSubdir(base.dir, subdir, ECD);
}

#endif /* KLUSTOR_H */
